# -*- coding: utf8 -*-
"""ICML Apprenticeship Atari Results

Automatically generated by coLaboratory.

Original file is located at
    https://drive.google.com/a/google.com/file/d/0B0ZDN_jTarLsOG1QeXlRc1oxYVU/view?usp=drivesdk

# Tensorflow DQN agent, Bigtable results

## A colab for visualizing DQN runs

Taken from the basic DQN colab. Adapted for DQN disney.
"""

#@title Imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

  
import os
import sys
import math
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import patsy

from itertools import cycle

import google3
from google3.pyglib import gfile
from google3.learning.deepmind.zephyr.python import util
from google3.learning.deepmind.xmanager.db import xmdb
from google3.learning.deepmind.xmanager import parameters

# deciding best for plots and tables
xmax = 200
best_window_length = 3
best_sampling_ratio = 0

#@title Bigtable read settings
owner = 'toddhester' #@param { isTemplate: true}
schema_name = "dqn31" #@param { isTemplate: true}
clock_name = "observer"
rowkeys="prefix,xid,wid,platform,game,seed"
experiment = "v1" #@param { isTemplate: true}

#@title Data Helper functions
def flatten(d, parent_key='', sep='.'):
  """Flattens a nested dictionary into a flat dict
  
  Args:
    d - dict - to be flattened
    parent_key - string
    sep - desired string separation
    
  Returns:
    flat_dict
  """
  items = []
  for k, v in d.items():
    new_key = parent_key + sep + k if parent_key else k
    if isinstance(v, dict):
      items.extend(flatten(v, new_key, sep=sep).items())
    else:
      items.append((new_key, v))
  return dict(items)


def xmdb_to_dataframe(xid, slicer=[]):
  """Loads in xmdb experiments as dataframe
  
  Args:
    xid - int - experiment id
    slicer - list of str - of config to remove + 'Parameters' automatic
  
  Returns:
    dt - pandas dataframe of results
  """
  db = xmdb.ExperimentManagerDB()
  work_units = []
  slicer = ['Parameters'] + slicer
  for wid in range(db.get_work_unit_count(xid)):
    print(wid)
    work_data = db.get_work_unit(xid, wid)
    for param in slicer:
      work_data = work_data[param]
    flat_work = flatten(work_data)
    flat_work['wid'] = wid
    work_units.append(flat_work)
#     work_data['wid'] = wid
#     work_units.append(work_data)
  
  return pd.DataFrame(flat_work)

def load_params(xid):
  """Loads in xmdb experiments as dataframe
  
  Args:
    xid - int - experiment id
    slicer - list of str - of config to remove + 'Parameters' automatic
  
  Returns:
    dt - pandas dataframe of results
  """
  dt_params = parameters.get_experiment_parameters(xid)
  dt_params.columns = map(lambda x: x.split('.')[-1], dt_params.columns)
  return dt_params

def clean_col_name(colname, lim='/'):
  """Clean an indiviual column name from zephyr
  
  Args:
    colname - string - from Zephyr, 'data/opt/batchsize/value'
      
  Returns:
    clean_name - string - like 'batchsize'
  """
  segments = colname.split(lim)
  if len(segments) == 1:
    return colname
  elif len(segments) == 2:
    return segments[1]
  elif len(segments) == 5:
    train_test = segments[1].split('_')[0]
    var_name = segments[len(segments) - 2]
    var_type = segments[len(segments) - 1]
    
    if var_type == 'value':
      clean_name = var_name + '_' + train_test
    else:
      clean_name = var_name + '_' + var_type + '_' + train_test
    return clean_name
  else:
    return colname

def clean_cols(data, lim='/'):
  """Clean the column names from zephyr
  
  Args:
    data - dataframe with zephyr colnames
  
  Returns:
    NULL - updated in place
  """
  new_names = []
  for colname in data.columns:
    new_names.append(clean_col_name(colname, lim=lim))
  data.columns = new_names
  

def load_catch_data(xid, owner='toddhester'):
  """Load in an entire xid data return clean dataframe.
  """
  
  # Xmanager parameters
  dt_params = parameters.get_experiment_parameters(xid)
  prefix = 'config.agent.'
  dt_params.columns = map(lambda x: '.'.join(x.split('.')[2:]), dt_params.columns)
  dt_params.reset_index(inplace=True)
  dt_params.rename(columns={'WId': 'wid'}, inplace=True)
  
  # Raw data from Xmanager
  dt_raw = util.ReadDataFrame(owner=owner, schema_name=schema_name, 
                              clock_name=clock_name, rowkeys=rowkeys,
                              select_rowkeys={"prefix": experiment, "xid": xid})
  dt_raw.reset_index(inplace=True)
  clean_cols(dt_raw)
  #print(dt_raw)
  dt_raw['xid'] = pd.to_numeric(dt_raw['xid'])
  dt_raw['wid'] = pd.to_numeric(dt_raw['wid'])
  
  hypers = set(dt_params.columns.values)
  hypers.remove('wid')
  hypers.remove('seed')
  
  return pd.merge(dt_raw, dt_params, how='left', on='wid'), hypers

#@title Plotting helper functions
def evaluate_every_dim(data, variables, averaged=None, 
                       yvariable='episode_returns_mean_test',
                       plotSingletons=False, window_length=5):

  # build dict of best value of each parameter
  best = {}
  
  variables = list(variables)
    
  # average across ticks, find variables for that best run.
  agg = data.reset_index()
  agg = agg.groupby('wid').mean()
  bestTask = agg['episode_returns_mean_train'].argmax()
  for variable in variables:
    bestValue = data.loc[bestTask][variable]
    try:
      best[variable] = bestValue.iloc[0]
    except:
      best[variable] = bestValue
  plotEachDim(data, best, yvariable, plotSingletons=plotSingletons, 
              window_length=window_length)
  
  
def getVaryingHypers(data, variables):
  variables = list(variables)
  
  # remove variables that do not change
  for variable in list(variables):
    if data[variable].nunique() == 1:
      variables.remove(variable)
  variables.append('xid')
  
  return variables
  
def plotEachDim(data, settings, yvariable, plotSingletons=False, 
                window_length=5):
  for variable in sorted(settings.keys()):
    # remove variable from set
    config = settings.copy()
    config.pop(variable, None)
    # filter data
    subset = filterToSubset(data, config)
    # plot
    #print(variable)
    #print(subset.index.unique())
    plotRewardAndMaxQ(subset, settings.keys(), variable, 'test', 
                      plotSingletons=plotSingletons, window_length=window_length)
    
    
def filterToSubset(data, filters):
  subset = data.copy()
  for key in filters.keys():
    #print('filter from {} on {} to {}'.format(len(data), key, filters[key]))
    subset = subset[(subset[key] == filters[key])]
    #print('new size {}'.format(len(subset)))
    
  # filter to the most common value of seed for the plots
  (values, counts) = np.unique(subset.seed_y, return_counts=True)
  index = np.argmax(counts)
  best_seed = values[index]
  subset = subset[(subset.seed_y == best_seed)]
  #print("Filter to seed value {}, final size {}".format(best_seed, len(subset)))

  return subset

#@title Plotting methods
def plotRewardAndMaxQ(parameters, variables, title='', subset='test', 
                      plotMaxQ=False, plotSingletons=False, window_length=5,
                     plotExpert=False):
  
  variables = list(variables)
  if 'seed_y' not in variables and 'seed_x' not in variables:
    variables.append('seed_y')
    
  if parameters.index.nunique() <= 1 and not plotSingletons:
    #print("only one parameter", title)
    return
  
  # if only seed varies but variable is not seed, dont plot
  if 'seed' not in title:
    something_varies = False
    for variable in variables:
      if parameters[variable].nunique() > 1 and 'seed' not in variable:
        something_varies = True
        break
    if not something_varies and not plotSingletons:
      #print("nothing varies", title)
      return
  
  # test reward and maxQ or train and test reward
  variable = 'step_maxQ_mean_{}'.format(subset)
  if True:
    fig, (ax1) = plt.subplots(1, 1, figsize=(8,8)) 
    fixed, ncolors = plotByTask(ax1, parameters, variables, 
                       title='Training Mean Rewards', 
                       ylabel='Mean Rewards Per Episode', 
                       yvariable='episode_returns_mean_train',
                       window_length=best_window_length, plotExpert=plotExpert)
    if True:
      if fixed.title().replace('_',' ') != title.title():
        title = '{}_{}'.format(title, fixed)
      #plt.savefig('{}_training_mean_rewards.svg'.format(title.replace(' ', '_')), format='svg', dpi=1000)
      plt.savefig('{}_training_mean_rewards.eps'.format(title.replace(' ', '_')), format='eps', dpi=1000)
  elif variable in parameters.columns and plotMaxQ:
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16,8)) 
    fixed = plotByTask(ax1, parameters, variables, 
                       title='{} {} Mean Rewards'.format(title, subset), 
                       ylabel='{} Episode Returns'.format(subset), 
                       yvariable='episode_returns_mean_{}'.format(subset),
                       window_length=window_length, plotExpert=plotExpert)
    plotByTask(ax2, parameters, title = '{} Max Q'.format(title), 
               ylabel = 'Max Q', yvariable=variable,
               window_length=window_length, plotExpert=plotExpert)
  else:
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(24,8)) 
    fixed = plotByTask(ax2, parameters, variables, 
                       title='{} Train Mean Rewards'.format(title), 
                       ylabel='Train Episode Returns', 
                       yvariable='episode_returns_mean_train',
                       window_length=window_length, plotExpert=plotExpert)
    #fixed = plotByTask(ax2, parameters, variables, 
    #                   title='{} Test Episode Steps'.format(title), 
    #                   ylabel='Test Episode Steps', 
    #                   yvariable='episode_steps_mean_test',
    #                   window_length=window_length, plotExpert=plotExpert)
    fixed = plotByTask(ax3, parameters, variables, 
                       title='{} Test # Rewards'.format(title), 
                       ylabel='Test # Rewards', 
                       yvariable='episode_num_rewards_mean_test',
                       window_length=window_length, plotExpert=plotExpert)
    plotByTask(ax1, parameters, variables, 
               title='{} Test Mean Rewards'.format(title), 
               ylabel='Test Episode Returns', 
               yvariable='episode_returns_mean_test',
               window_length=window_length, plotExpert=plotExpert)
  
  fig.suptitle(title)
  #plt.show()
  
  
def abbreviate(text):
  if 'heads' in text:
    return 'heads'
  if 'sstable' in text:
    return 'log'
  if text == 'max_gamma_schedule.start_val':
    return 'maxgmma'
  if text == 'epsilon_train_schedule.start_val':
    return 'epsilon'
  if text == 'update_target':
    return 'tgt'
  if text == 'argmax_values':
    return 'argmax'
  if text == 'action_selection':
    return 'act'
  if text == 'argmax_std_penalty':
    return 'argmaxpen'
  if text == 'target_std_penatly':
    return 'tgtpen'
  if text == 'action_std_penalty':
    return 'actpen'
  if text == 'noise_scale':
    return 'noise'
  if text == 'expert_margin':
    return 'margin'
  if text == 'expert_weight_schedule.start_val':
    return 'exp wt start'
  if text == 'expert_weight_schedule.end_val':
    return 'exp wt end'
  if text == 'sampling_ratio_schedule.start_val':
    return 'sampling start'
  if text == 'sampling_ratio_schedule.end_val':
    return 'sampling end'
  if text == 'expert_weight':
    return 'exp wt'
  if text == 'arguments.levelName':
    return 'level'
  if text == 'arguments.mode':
    return 'mode'
  if text == 'expert_sampling_ratio':
    return 'sample ratio'
  if text == 'two_buffers':
    return '2buffers'
  if text == 'data_addition_schedule.end_time':
    return 'pretrain steps'
  words = text.split('_')
  return ''.join([w[:3] if len(w) > 4 else w for w in words])
  #vowels = ['a','e','i','o','u','_']
  #return ''.join([l for l in text if l not in vowels])
  

def getSpecialLabel(params, ncolors, nreg, nratios):
  if params.load_replay_buffer_from_sstable == "Learn From Scratch":
    if params.max_abs_reward == 1:
      return "Learn From Scratch, Clipped Rewards"
    else:
      return "Double DQN"
  if params.pretraining_iterations == 0:
    return "No Pre-training"
  if params.expert_weight == 0:
    return "No Supervised Loss"
  if params.td_weight == 0:
    return "No Q-Learning Loss"
  if params.regularization_weight == 0:
    return "No L2 regularization"
  if params.max_abs_reward == 1:
    return "Clipped Rewards"
  if params['epsilon_train_schedule.start_val'] > 0.01:
    return "Epsilon=0.2 Exploration"
  if params.expert_error_type == 'cross_entropy':
    return "Cross Entropy Loss"
  if params.expert_bonus > 0.0:
    return "DQfD: With Expert Bonus"
  if ncolors > 3 and nratios > 2:
    if params.expert_sampling_ratio == 0.8 or params['sampling_ratio_schedule.start_val'] == 0.8:
      return "DQfD: 0.8 Sampling Ratio"
    if params['sampling_ratio_schedule.start_val'] == 0.5:
      return "DQfD: 0.5 Sampling Ratio"
    if params['sampling_ratio_schedule.start_val'] == 0.1:
      return "DQfD: 0.1 Sampling Ratio"
    if params['sampling_ratio_schedule.start_val'] == 1.0:
      if params['sampling_ratio_schedule.end_time'] == 250000 * 100:
        return "Decay over 100 iterations"
      else:
        return "Decay over 50 iterations"
  return "DQfD"


def plotByTask(ax, parameters, variables, title='test', ylabel='Epsiode Returns', 
               yvariable='episode_returns_mean_test', window_length=5, 
               plotExpert=True):
  
  # set larger color range
  NUM_COLORS = len(parameters.index.unique())
  
  if plotExpert:
    NUM_COLORS += 1
  
  #cm = plt.get_cmap('gist_rainbow')
  #ax.set_color_cycle([cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)])
  tasks = sorted(parameters.index.unique())
  
  lines = ['-','--',':','-.']
  linecycler = cycle(lines) 
    
  for task in tasks:
    subset = parameters[(parameters.index == task)]
    first = subset.iloc[0]
    if len(subset) < window_length:
      #print("length is less than window", window_length, title)
      continue
    label = ''
    fixedparams = ''

    nvariables = 0
    for variable in variables:
      if variable in parameters.columns:
        if parameters[variable].nunique() > 1:
          #label += '{}, '.format(first[variable])
          label +=  '{} {}, '.format(first[variable], abbreviate(variable))
          nvariables += 1
        else:
          fixedparams += '{} {}, '.format(first[variable], abbreviate(variable))

    # special label for ones with lots of variables
    if nvariables > 0:
      label = getSpecialLabel(first, NUM_COLORS, parameters.regularization_weight.nunique(), parameters['sampling_ratio_schedule.start_val'].nunique())
      fixedparams = first['arguments.levelName']
    else:
      label = label[:-2] + ", {}".format(task)
      fixedparams = fixedparams[:-2]
    #print(title, label, subset[yvariable].unique(), subset.tick.unique())
    data = subset.rolling(window=window_length,center=False).mean()
    if label == 'DQfD' or label == 'DQfD: 0.1 Sampling Ratio':
      color = 'b'
      linestyle = '-'
    elif label == 'DQfD: 0.5 Sampling Ratio':
      color = 'c'
      linestyle = '-.'
    elif label == 'DQfD: 0.8 Sampling Ratio':
      color = 'm'
      linestyle = ':'
    elif label == 'No Pre-training':
      color = 'c'
      linestyle = '-.'
    elif label == 'Cross Entropy Loss':
      color = 'c'
      linestyle = '-'
    elif label == 'No L2 regularization':
      color = 'm'
      linestyle = '-.'
    elif label == 'No Q-Learning Loss':
      color = 'y'
      linestyle = ':'
    elif label == 'No Supervised Loss':
      color = 'k'
      linestyle = '--'
    elif label == 'Double DQN':
      color = 'g'
      linestyle = '--'
    else:
      color = next(ax._get_lines.prop_cycler)['color']
      linestyle = next(linecycler)
    ax.plot(data['tick'], data[yvariable], label=label, linewidth=3, color=color, linestyle=linestyle)
    
  # add expert data
  if False:
    expert_data = expert_scores[(expert_scores.Game == first['arguments.levelName'])]
    for variable in ['Min', 'Max', 'Mean']:
      value = expert_data[variable].iloc[0]
      color = next(ax._get_lines.prop_cycler)['color']
      ax.axhline(y=value, label='Expert {}'.format(variable), color=color, linewidth=3, linestyle=next(linecycler))
  
  # imitation
  if plotExpert:
    color = 'r'
    linestyle = ':'
    #color = next(ax._get_lines.prop_cycler)['color']
    ax.axhline(y=imitation[first['arguments.levelName']], label='Imitation', color=color, linewidth=3, linestyle=linestyle)
      
  ax.set_xlim([0,xmax])
  ax.set_ylabel(ylabel)
  ax.set_xlabel('Training Iteration')
  handles, labels = ax.get_legend_handles_labels()
  # sort both labels and handles by labels
  if len(labels) > 1:
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
  ax.legend(handles, labels, loc=0)
  #ax.legend()
  #ax.set_title(title) #'{} ({})'.format(title, fixedparams))
  ax.set_title(first['arguments.levelName'].title().replace('_',' ')) #'{} ({})'.format(title, fixedparams))

  return fixedparams, NUM_COLORS

# load expert and imitation scores
with gfile.Open('/cns/wj-d/home/deepmind/modeldata/apprenticeship/atari_expert_scores.csv') as filename:
  expert_scores = pd.read_csv(filename)


#@title Lineup pre-training phases

def compute_pretraining_offset(x):
  if 'pretraining_iterations' in x:
    return x.tick - x.pretraining_iterations
  if 'data_addition_schedule.end_time' in x:
    return x.tick - x['data_addition_schedule.end_time']/250000
  elif x.load_replay_buffer_from_sstable == 'Learn From Scratch':
    return x.tick
  else:
    return x.tick-100

def dataframe_fixes(dt):
  # rename no log to 'learn from scratch'
  dt.load_replay_buffer_from_sstable = dt.apply(lambda x: x.load_replay_buffer_from_sstable.split('/')[-1] if len(x.load_replay_buffer_from_sstable) > 0 else "Learn From Scratch", axis=1)

  # Subtract pre-training iterations so iteration 0 is always first interactions
  dt.tick = dt.apply(compute_pretraining_offset, axis=1)

  # convert list hypers to tuples
  list_hypers = ['observation_shape', 'convnet_args.n_hidden']
  for hyper in list_hypers:
    if hyper in hypers:
      dt[hyper] = dt.apply(lambda x: tuple(x[hyper]), axis=1)

  # replace NaN's
  dt.fillna(value=0, inplace=True)
  
  return dt
    
def average_dataframe(dt, hypers):
  #@title Average over seeds
  averaged = dt.reset_index()

  gb_cols = list(hypers)
  gb_cols.append('tick')
  gb_cols.remove('epsilon_evaluation')
  gb_cols.remove('test_frames')
  averaged = averaged.groupby(gb_cols).mean()
  averaged = averaged.reset_index().set_index('wid')

        
  # replace training data from before tick 0 with Nones rather than 0s
  for item in averaged.columns:
    if 'epsilon_train_schedule' in item:
      continue
    if '_train' in item:
      averaged.loc[averaged.tick < 0, item] = None
  
  # get changing hypers
  changed_hypers = getVaryingHypers(averaged, hypers)
  return averaged, changed_hypers

def get_final_policy_values(averaged):
  imitation = {}
  for game in sorted(averaged['arguments.levelName'].unique()): 
    data = averaged[(averaged['arguments.levelName'] == game)]

    # get policy value averaged over last 20 iterations
    #rolling = pd.rolling_mean(data, 20)
    rolling = data.rolling(window=20, center=False, min_periods=1).mean()
    last = rolling[(rolling.tick == rolling.tick.max())]
    #print(game, last)
    #'Final Test:', last.episode_returns_mean_test.iloc[0], 'Final Train:', last.episode_returns_mean_train.iloc[0])
    imitation[game] = last.episode_returns_mean_test.iloc[0]
  return imitation

# load imitation data
dt1, hypers1 = load_catch_data(13582)
dt1 = dt1.reset_index().set_index('wid')

dt2, hypers2 = load_catch_data(14710)
dt2 = dt2.reset_index().set_index('wid')

dt3, hypers3 = load_catch_data(15022)
dt3 = dt3.reset_index().set_index('wid')

dt2.index = dt2.index + 1000
dt3.index = dt3.index + 2000

dt = pd.concat([dt1, dt2, dt3])

hypers = list(set(hypers1) | set(hypers2) | set(hypers3))
dt = dataframe_fixes(dt)
averaged, changed_hypers = average_dataframe(dt, hypers)
imitation = get_final_policy_values(averaged)

def load_and_concat_dataframes(xids):
  dataframes = []
  hypers = set()
  for i, item in enumerate(xids):
    if isinstance(item, list):
      xid = item[0]
      user = item[1]
    else:
      xid = item
      user = 'toddhester'
    print('loading', xid, user)
    df, h = load_catch_data(xid, user)
    # make sure they all have unique wid's
    df.wid = df.wid + i * 1000
    dataframes.append(df)
    hypers = hypers | h
    
  # merge
  fulldt = pd.concat(dataframes)
  
  # reset index
  fulldt = fulldt.reset_index().set_index('wid')
  
  return fulldt, hypers

fulldt, hypers = load_and_concat_dataframes([13442, 13975, 13986, 13979, 13982, 14712, [14971, 'matejvecerik'], 14982, 15255, 15379, 15391, 15390, 16359])

dt = dataframe_fixes(fulldt)
averaged, changed_hypers = average_dataframe(dt, hypers)

"""## Plotting learning curves"""

for game in sorted(averaged['arguments.levelName'].unique()): 
  data = averaged[(averaged['arguments.levelName'] == game)]
  data = data.reset_index()
  #print(game)

  
  x_hypers = list(changed_hypers)

  #(values, counts) = np.unique(data.seed_y, return_counts=True)
  #index = np.argmax(counts)
  #best_seed = values[index]
  #subset = data[(data.seed_y == best_seed)]
  subset = data

  # pull out particular runs
  scratch = subset[(subset.load_replay_buffer_from_sstable == "Learn From Scratch") & (subset.max_abs_reward == 100000)]
  full = subset[(subset.load_replay_buffer_from_sstable != "Learn From Scratch") & (subset.pretraining_iterations > 0) & (subset.expert_error_type == 'large_margin') & (subset['epsilon_train_schedule.start_val'] == 0.01) & (subset.max_abs_reward == 100000) & (subset.regularization_weight > 0) & (subset.td_weight > 0) & (subset.expert_weight > 0)]
  nobonus = subset[(subset.load_replay_buffer_from_sstable != "Learn From Scratch") & (subset.pretraining_iterations > 0) & (subset.expert_error_type == 'large_margin') & (subset['epsilon_train_schedule.start_val'] == 0.01) & (subset.max_abs_reward == 100000) & (subset.regularization_weight > 0) & (subset.expert_bonus == 0) & (subset.td_weight > 0) & (subset.expert_weight > 0)]
  nobonus = nobonus[(nobonus['sampling_ratio_schedule.start_val'] == 0.1)]

  samplings = {}
  samplings[0.8] = full[((full.expert_sampling_ratio == 0.8) | (full['sampling_ratio_schedule.start_val'] == 0.8)) & (full.expert_bonus == 0)]
  samplings[0.5] = full[(full['sampling_ratio_schedule.start_val'] == 0.5) & (full.expert_bonus == 0)]
  samplings[0.1] = full[(full['sampling_ratio_schedule.start_val'] == 0.1) & (full.expert_bonus == 0)]
  samplings[0] = nobonus

  # reset indices on these for plotting
  scratch.wid = 0
  samplings[0.1].wid = -100
  samplings[0.5].wid = -50
  samplings[0.8].wid = -8
  samplings[0].wid = -11

  # try different sizes
  plt.rcParams.update({'axes.titlesize': 'xx-large',
                       'axes.labelsize': 'medium',
                       'xtick.labelsize': 'medium',
                       'ytick.labelsize': 'medium',
                       'legend.fontsize': 'large'})
    
  # plot all RL+LfD options
  if game == 'freeway':
    both = pd.concat([scratch, samplings[0.1], samplings[0.5], samplings[0.8]])
    both = both.set_index('wid')
    plotRewardAndMaxQ(both, x_hypers, window_length=1, plotExpert=True, title="Sampling Ratios")

  # plot with and without expert bonus
  #both = pd.concat([scratch, samplings[0.1], samplings[0]])
  #both = both.set_index('wid')
  #plotRewardAndMaxQ(both, x_hypers, window_length=1, plotExpert=False, title="Expert Bonus")
  
  # plot a single LfD option
  # choose which one to plot
  comparison = samplings[best_sampling_ratio]

  plt.rcParams.update({'axes.titlesize': 'xx-large',
                       'axes.labelsize': 'large',
                       'xtick.labelsize': 'large',
                       'ytick.labelsize': 'large',
                       'legend.fontsize': 'x-large'})
  both = pd.concat([scratch, comparison])
  both = both.set_index('wid')
  plotRewardAndMaxQ(both, x_hypers, window_length=1, plotExpert=True, title=game.title().replace('_',' '))
  

for game in ['private_eye', 'road_runner']: 
  data = averaged[(averaged['arguments.levelName'] == game)]
  #(values, counts) = np.unique(data.seed_y, return_counts=True)
  #index = np.argmax(counts)
  #best_seed = values[index]
  #subset = data[(data.seed_y == best_seed)]
  subset = data.reset_index()
  
  #print(game)
  
  x_hypers = list(changed_hypers)

  sdata = {}
  sdata[0.1] = subset[(subset['sampling_ratio_schedule.start_val'] == 0.1) & (subset.expert_bonus > 0)]
  sdata[0.8] = subset[(subset.expert_sampling_ratio == 0.8) | (subset['sampling_ratio_schedule.start_val'] == 0.8) & (subset.expert_bonus > 0)]
  sdata[0.5] = subset[(subset['sampling_ratio_schedule.start_val'] == 0.5) & (subset.expert_bonus > 0)]
  sdata[0] = subset[(subset['sampling_ratio_schedule.start_val'] == 0.1) & (subset.expert_bonus == 0)]
  

  # different sampling rates
  for sampling_ratio in [best_sampling_ratio]:
  
    subdata = sdata[sampling_ratio]    
    plt.rcParams.update({'axes.titlesize': 'xx-large',
                       'axes.labelsize': 'medium',
                       'xtick.labelsize': 'medium',
                       'ytick.labelsize': 'medium',
                       'legend.fontsize': 'large'})
  
    # Reward clipping plot
    anyscratch = subset[(subset.load_replay_buffer_from_sstable == 'Learn From Scratch')]
    full = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.pretraining_iterations > 0) & (subdata.expert_error_type == 'large_margin') & (subdata['epsilon_train_schedule.start_val'] == 0.01) & (subdata.max_abs_reward == 100000) & (subdata.regularization_weight > 0) & (subdata.td_weight > 0) & (subdata.expert_weight == 1.0)]
    clipped = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.pretraining_iterations > 0) & (subdata.expert_error_type == 'large_margin') & (subdata['epsilon_train_schedule.start_val'] == 0.01) & (subdata.max_abs_reward == 1) & (subdata.regularization_weight > 0) & (subdata.td_weight > 0) & (subdata.expert_weight > 0)]
    data2 = pd.concat([anyscratch, full, clipped])
    #plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=False)

    # Pre-training
    scratch = subset[(subset.load_replay_buffer_from_sstable == 'Learn From Scratch') & (subset.max_abs_reward == 100000)]
    nopretrain = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.pretraining_iterations == 0)]
    scratch.wid = 0
    nopretrain.wid = -1
    full.wid = -2
    data2 = pd.concat([full, nopretrain, scratch])
    data2 = data2.set_index('wid')
    
    plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=True, title='Pre-Training Comparison')
    
    # Losses
    no_td = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.td_weight == 0)]
    no_expert = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.expert_weight == 0)]
    no_l2 = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.regularization_weight == 0)]
    cross_entropy = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.expert_error_type == 'cross_entropy')]
    data2 = pd.concat([full, no_td, no_expert, no_l2])
    #plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=True, title="Losses")
    
    # Cross entropy
    data2 = pd.concat([full, cross_entropy])
    #plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=True, title="Cross Entropy v Large Margin")

    # losses and cross entropy
    full.wid = -10
    no_td.wid = -9
    no_expert.wid = -8
    no_l2.wid = -7
    cross_entropy.wid = -6
    scratch.wid = -5
    data2 = pd.concat([full, no_td, no_expert, no_l2, cross_entropy])
    data2 = data2.set_index('wid')
    #data2 = pd.concat([full, no_td, no_expert, no_l2, cross_entropy])
    plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=True, title="Losses2")
    
    # Bonus
    no_bonus = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata.expert_bonus == 0)]
    data2 = pd.concat([full, no_bonus])
    #plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=False, title="Expert Bonus")
    
    # exploration
    high_epsilon = subdata[(subdata.load_replay_buffer_from_sstable != "Learn From Scratch") & (subdata['epsilon_train_schedule.start_val'] > 0.01)]
    data2 = pd.concat([full, high_epsilon])
    #plotRewardAndMaxQ(data2, x_hypers, window_length=1, plotExpert=False, title="Exploration")
    
    # all plots
    #plotRewardAndMaxQ(subdata, x_hypers, window_length=1, plotExpert=False, title="Full Comparison")

def fix_game_name(game):
  game = game.title().replace('_',' ')
  if game == 'Jamesbond':
    return 'James Bond'
  if game == 'Montezuma Revenge':
    return "Montezuma's Revenge"
  if game == "Riverraid":
    return "River Raid"
  if game == "Qbert":
    return "Q-bert"
  if game == "Yars Revenge":
    return "Yars' Revenge"
  return game
  
# table data
for ratio in [0, 0.1, 0.5, 0.8]:
  total_games = 0
  total_dqfd = 0
  total_dqn = 0
  total_imitation = 0
  dqfd_best = 0
  dqn_best = 0
  imitation_best = 0
  final_dqfd_best = 0
  final_dqn_best = 0
  final_imitation_best = 0
  initial_dqfd_best = 0
  initial_dqn_best = 0
  initial_imitation_best = 0
  cum_dqfd_over_dqn = 0
  cum_dqfd_over_imit = 0
  final_dqfd_over_dqn = 0
  final_dqfd_over_imit = 0
  initial_dqfd_over_dqn = 0
  initial_dqfd_over_imit = 0
  dqn_better_than_dqfd = []
  dqn_better_initially = []
  better_than_expert_min = []
  better_than_expert_max = []
  print('\nTable for sampling ratio {}'.format(ratio))

  for game in sorted(averaged['arguments.levelName'].unique()): 
    data = averaged[(averaged['arguments.levelName'] == game)]
    data = data.reset_index()
    #print(game)

    x_hypers = list(changed_hypers)

    #(values, counts) = np.unique(data.seed_y, return_counts=True)
    #index = np.argmax(counts)
    #best_seed = values[index]
    #subset = data[(data.seed_y == best_seed)]
    subset = data

    # pull out particular runs
    scratch = subset[(subset.load_replay_buffer_from_sstable == "Learn From Scratch") & (subset.max_abs_reward == 100000)]
    full = subset[(subset.load_replay_buffer_from_sstable != "Learn From Scratch") & (subset.pretraining_iterations > 0) & (subset.expert_error_type == 'large_margin') & (subset['epsilon_train_schedule.start_val'] == 0.01) & (subset.max_abs_reward == 100000) & (subset.regularization_weight > 0) & (subset.td_weight > 0) & (subset.expert_weight > 0)]
    nobonus = subset[(subset.load_replay_buffer_from_sstable != "Learn From Scratch") & (subset.pretraining_iterations > 0) & (subset.expert_error_type == 'large_margin') & (subset['epsilon_train_schedule.start_val'] == 0.01) & (subset.max_abs_reward == 100000) & (subset.regularization_weight > 0) & (subset.expert_bonus == 0) & (subset.td_weight > 0) & (subset.expert_weight > 0)]
    nobonus = nobonus[(nobonus['sampling_ratio_schedule.start_val'] == 0.1)]
    samplings = {}
    samplings[0.8] = full[(full.expert_sampling_ratio == 0.8) | (full['sampling_ratio_schedule.start_val'] == 0.8) & (full.expert_bonus > 0)]
    samplings[0.5] = full[(full['sampling_ratio_schedule.start_val'] == 0.5) & (full.expert_bonus > 0)]
    samplings[0.1] = full[(full['sampling_ratio_schedule.start_val'] == 0.1) & (full.expert_bonus > 0)]
    samplings[0] = nobonus

    # reset indices on these for plotting
    scratch.wid = 0
    samplings[0.1].wid = -1
    samplings[0.5].wid = -5
    samplings[0.8].wid = -8
    samplings[0].wid = -11

    comparison = samplings[ratio]

    #for ratio in samplings.keys():
    # print cumulative rewards
    #comparison = samplings[ratio]
    num_iterations = xmax
    min_iter_complete = min(comparison.tick.max(), scratch.tick.max())
    num_iterations = min(xmax, min_iter_complete)
    cum_dqfd = comparison[(comparison.tick < num_iterations)].episode_returns_mean_train.mean()
    cum_dqn = scratch[(scratch.tick < num_iterations)].episode_returns_mean_train.mean()
    cum_imitation = imitation[game]
    if cum_dqfd > cum_dqn and cum_dqfd > cum_imitation:
      print('{} & \\textbf{{{:.1f}}} & {:.1f} & {:.1f} \\\\ % {} iterations'.format(fix_game_name(game), cum_dqfd, cum_dqn, cum_imitation, num_iterations))
    elif cum_dqn > cum_dqfd and cum_dqn > cum_imitation:
      print('{} & {:.1f} & \\textbf{{{:.1f}}} & {:.1f} \\\\ % {} iterations'.format(fix_game_name(game), cum_dqfd, cum_dqn, cum_imitation, num_iterations))
    else:
      print('{} & {:.1f} & {:.1f} & \\textbf{{{:.1f}}} \\\\ % {} iterations'.format(fix_game_name(game), cum_dqfd, cum_dqn, cum_imitation, num_iterations))
      total_games += 1

    # cumulative reward stats
    if cum_dqfd > cum_dqn:
      cum_dqfd_over_dqn += 1
    else:
      dqn_better_than_dqfd.append(game)
    if cum_dqfd > cum_imitation:
      cum_dqfd_over_imit += 1
    if cum_dqfd > cum_dqn and cum_dqfd > cum_imitation:
      dqfd_best += 1
    elif cum_dqn > cum_dqfd and cum_dqn > cum_imitation:
      dqn_best += 1
    else:
      imitation_best += 1
      total_dqfd += cum_dqfd/ float(num_iterations)
      total_dqn += cum_dqn/ float(num_iterations)
      total_imitation += cum_imitation/ float(num_iterations)

    # final policies
    rolling = comparison[(comparison.tick < num_iterations)].rolling(window=10, center=False, min_periods=1).mean()
    dqfd_final = rolling[(rolling.tick == rolling.tick.max())].episode_returns_mean_train.iloc[0]
    rolling = scratch[(scratch.tick < num_iterations)].rolling(window=10, center=False, min_periods=1).mean()
    dqn_final = rolling[(rolling.tick == rolling.tick.max())].episode_returns_mean_train.iloc[0]

    if dqfd_final > dqn_final and dqfd_final > imitation[game]:
      final_dqfd_best += 1
    elif dqn_final > dqfd_final and dqn_final > imitation[game]:
      final_dqn_best += 1
    else:
      final_imitation_best += 1

    if dqfd_final > dqn_final:
      final_dqfd_over_dqn += 1
    if dqfd_final > imitation[game]:
      final_dqfd_over_imit += 1

    # initial policies
    dqfd_initial = comparison[(comparison.tick == 0)].episode_returns_mean_train.iloc[0]
    dqn_initial = scratch[(scratch.tick == 0)].episode_returns_mean_train.iloc[0]

    if dqfd_initial > dqn_initial and dqfd_initial > imitation[game]:
      initial_dqfd_best += 1
    elif dqn_initial > dqfd_initial and dqn_initial > imitation[game]:
      initial_dqn_best += 1
    else:
      initial_imitation_best += 1

    if dqfd_initial > dqn_initial:
      initial_dqfd_over_dqn += 1
    else:
      dqn_better_initially.append(game)
    if dqfd_initial > imitation[game]:
      initial_dqfd_over_imit += 1

    expert_data = expert_scores[(expert_scores.Game == game.title().replace('_',' '))]
    min_expert = expert_data['Min'].iloc[0]
    max_expert = expert_data['Max'].iloc[0]
    if dqfd_final > min_expert:
      better_than_expert_min.append(game)
    if dqfd_final > max_expert:
      better_than_expert_max.append(game)

  print(' ')
  print('xmax:', xmax, 'Ratio:', ratio)
  print('Games:', total_games, 'DQFD best cumreward:', dqfd_best, 'DQN best cumreward:', dqn_best, 'imit best cumreward:', imitation_best, 'DQFD > DQN:', cum_dqfd_over_dqn, 'DQFD > imit:', cum_dqfd_over_imit)
  print("DQN better in cumulative rewards:", dqn_better_than_dqfd)
  print('DQFD best final:', final_dqfd_best, 'DQN best final:', final_dqn_best, 'imit best final:', final_imitation_best, 'DQFD > DQN:', final_dqfd_over_dqn, 'DQFD > imit:', final_dqfd_over_imit)
  print('Better than worst expert episode', len(better_than_expert_min), better_than_expert_min)
  print('Better than best expert episode', len(better_than_expert_max), better_than_expert_max)
  print('DQFD best initial:', initial_dqfd_best, 'DQN best initial:', initial_dqn_best, 'imit best initial:', initial_imitation_best, 'DQFD > DQN:', initial_dqfd_over_dqn, 'DQFD > imit:', initial_dqfd_over_imit)
  print('DQN better initially:', dqn_better_initially)
  print(total_dqfd, total_dqn, total_imitation)
  print(' ')
  
